PFX_TOKEN_SHORT_VALIDITY = {'hours': 12}
PFX_TOKEN_LONG_VALIDITY = {'days': 30}
PFX_TOKEN_OTP_VALIDITY = {'minutes': 15}
PFX_HOTP_CODE_VALIDITY = 15
PFX_OTP_VALID_WINDOW = 1
PFX_OTP_IMAGE = None
PFX_OTP_COLOR = None

PFX_COOKIE_DOMAIN = None
PFX_COOKIE_SECURE = True
PFX_COOKIE_SAMESITE = 'None'

PFX_LOGIN_BAN_FAILED_NUMBER = 5
PFX_LOGIN_BAN_SECONDS_START = 60
PFX_LOGIN_BAN_SECONDS_STEP = 60

PFX_OPENAPI_PATH = "doc/api/"
PFX_OPENAPI_FILENAME = "openapi"
PFX_OPENAPI_TEMPLATE = {}

PFX_PROFILING_SQL = ""
PFX_PROFILING_CPROFILE = ""

STORAGE_DEFAULT = None

STORAGE_S3_AWS_REGION = None
STORAGE_S3_AWS_ACCESS_KEY = None
STORAGE_S3_AWS_SECRET_KEY = None
STORAGE_S3_AWS_ENDPOINT_URL = None
STORAGE_S3_AWS_S3_BUCKET = None
STORAGE_S3_AWS_PUT_URL_EXPIRE = None
STORAGE_S3_AWS_GET_URL_EXPIRE = None

STORAGE_LOCAL_ROOT = None
STORAGE_LOCAL_X_ACCEL_REDIRECT = False

PFX_TEST_MODE = False

PFX_AUTH_GROUPS_CREATE_ONLY = False
