from .basic_api_errors import BasicAPIErrorTest
from .basic_api_test import BasicAPITest
from .test_api_doc import ApiDocTest
from .test_api_doc_search import TestAPIDocSearch
from .test_auth_api import AuthAPITest
from .test_body_mixin import TestBodyMixin
from .test_cache import TestCache
from .test_client import TestApiClient
from .test_fields_choices import TestFieldsChoices
from .test_fields_date import TestFieldsDate
from .test_fields_decimal import TestFieldsDecimal
from .test_fields_minutes_duration import TestFieldsMinutesDuration
from .test_fields_one2many import TestFieldsOne2Many
from .test_fields_rich_text import TestFieldsRichText
from .test_filters import FiltersTest
from .test_locale_api import LocaleAPITest
from .test_ordered_rest_view_mixin import TestOrderedRestViewMixin
from .test_perm_tests import PermTestsTest
from .test_permissions import TestPermissions
from .test_perms_api import PermsAPITest
from .test_post_migrate_groups_update import TestPostMigrateGroupsUpdate
from .test_profiling_middleware import ProfilingMiddlewareTest
from .test_settings import TestSettings
from .test_shortcuts import ShortcutTest
from .test_timezone_middleware import TimezoneMiddlewareTest
from .test_tools import TestTools
from .test_user_queryset import TestUserQuerySet
from .test_view_decorators import ViewDecoratorTest
from .test_view_fields import ViewFieldTest
